import io
from PIL import Image
import torch
from torch.utils.data import DataLoader
from torchvision import transforms
import random
import re


class FilePathLoader:
    def __init__(self, dataset, file_path_column="file_path", label_column=None):
        self.dataset = dataset
        self.file_path_column = file_path_column
        self.label_column = label_column
        self.successful_count = 0
        self.skipped_count = 0
        
        if self.file_path_column not in dataset.column_names:
            raise ValueError(f"Column '{self.file_path_column}' not found in dataset.")
            
    def _get_image(self, item):
        try:
            path = item[self.file_path_column]
            image = Image.open(path).convert("RGB")
            return image
        except Exception as e:
            print(f"Error loading image from {path}: {e}")
            return None
    
    def get_item(self, idx):
        item = self.dataset[idx]
        image = self._get_image(item)
        if image is not None:
            self.successful_count += 1
            if self.label_column is not None:
                label = item.get(self.label_column)
                return (image, label)
            else:
                return (image,)
        else:
            self.skipped_count += 1
            return None
    
    def print_summary(self):
        print(f"Successfully converted {self.successful_count} images.")
        print(f"Skipped {self.skipped_count} images due to errors.")

class JPGImageLoader:
    def __init__(self, dataset, bytes_column="jpg", label_column=None):
        self.dataset = dataset
        self.bytes_column = bytes_column
        self.label_column = label_column
        self.successful_count = 0
        self.skipped_count = 0
        
        if self.bytes_column not in dataset.column_names:
            raise ValueError(f"Column '{self.bytes_column}' not found in dataset.")
            
    def _get_image(self, item):
        try:
            bytes_data = item[self.bytes_column]
            img_bytes = io.BytesIO(bytes_data)
            image = Image.open(img_bytes).convert("RGB")
            return image
        except Exception as e:
            print(f"Error loading image from bytes: {e}")
            return None
    
    def get_item(self, idx):
        item = self.dataset[idx]
        image = self._get_image(item)
        if image is not None:
            self.successful_count += 1
            if self.label_column is not None:
                label = item.get(self.label_column)
                return (image, label)
            else:
                return (image,)
        else:
            self.skipped_count += 1
            return None
    
    def print_summary(self):
        print(f"Successfully converted {self.successful_count} images.")
        print(f"Skipped {self.skipped_count} images due to errors.")
        

class AIIADataLoader(DataLoader):
    def __init__(self, dataset,
                 batch_size=32,
                 val_split=0.2,
                 seed=42,
                 column="file_path",
                 label_column=None):
        super().__init__()
        
        self.batch_size = batch_size
        self.val_split = val_split
        self.seed = seed
        
        # Determine which loader to use based on the dataset's content
        # Check if any entry in bytes_column is a bytes or bytestring type
        is_bytes_or_bytestring = any(
            isinstance(value, (bytes, memoryview)) 
            for value in dataset[column].dropna().head(1).astype(str)
        )
        
        if is_bytes_or_bytestring:
            self.loader = JPGImageLoader(
                dataset,
                bytes_column=column,
                label_column=label_column
            )
        else:
            # Check if file_path column contains valid image file paths (at least one entry)
            sample_paths = dataset[column].dropna().head(1).astype(str)
            
            # Regex pattern for matching image file paths (adjust as needed)
            filepath_pattern = r'.*(?:/|\\).*\.([jJ][pP][gG]|png|gif)$'
            
            if any(
                re.match(filepath_pattern, path, flags=re.IGNORECASE) 
                for path in sample_paths
            ):
                self.loader = FilePathLoader(
                    dataset,
                    file_path_column=column,
                    label_column=label_column
                )
            else:
                # If neither condition is met, default to JPGImageLoader (assuming bytes are stored as strings)
                self.loader = JPGImageLoader(
                    dataset,
                    bytes_column=column,
                    label_column=label_column
                )
        
        # Get all items
        self.items = [self.loader.get_item(idx) for idx in range(len(dataset))]
        
        # Split into train and validation sets
        train_indices, val_indices = self._split_data()
        
        # Create datasets for training and validation
        self.train_dataset = self._create_subset(train_indices)
        self.val_dataset = self._create_subset(val_indices)
        
    def _split_data(self):
        if len(self.items) == 0:
            return [], []
            
        tasks = [item[1] if len(item) > 1 and hasattr(item, '__getitem__') else None for item in self.items]
        unique_tasks = list(set(tasks)) if tasks.count(None) < len(tasks) else []
        
        train_indices = []
        val_indices = []
        
        for task in unique_tasks:
            task_indices = [i for i, t in enumerate(tasks) if t == task]
            n_val = int(len(task_indices) * self.val_split)
            
            random.shuffle(task_indices)
            
            val_indices.extend(task_indices[:n_val])
            train_indices.extend(task_indices[n_val:])
            
        return train_indices, val_indices
    
    def _create_subset(self, indices):
        subset_items = [self.items[i] for i in indices]
        return AIIADataset(subset_items)
    
class AIIADataset(torch.utils.data.Dataset):
    def __init__(self, items):
        self.items = items
        
    def __len__(self):
        return len(self.items)
        
    def __getitem__(self, idx):
        item = self.items[idx]
        if isinstance(item, tuple) and len(item) == 2:
            image, label = item
            return (image, label)
        elif isinstance(item, tuple) and len(item) == 3:
            image, task, label = item
            # Handle tasks accordingly (e.g., apply different augmentations)
            if task == 'denoise':
                noise_std = 0.1
                noisy_img = image + torch.randn_like(image) * noise_std
                target = image
                return (noisy_img, target, task)
            elif task == 'rotate':
                angles = [0, 90, 180, 270]
                angle = random.choice(angles)
                rotated_img = transforms.functional.rotate(image, angle)
                target = torch.tensor(angle).long()
                return (rotated_img, target, task)
            else:
                raise ValueError(f"Unknown task: {task}")
        else:
            # Handle single images without labels or tasks
            if isinstance(item, Image.Image):
                return item
            else:
                raise ValueError("Invalid item format.")
