import torch
from torch import nn
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms
from PIL import Image
import os
import random
import pandas as pd
from aiia.model.config import AIIAConfig
from aiia.model import AIIABase
from aiia.data.DataLoader import AIIADataLoader

def pretrain_model(data_path1, data_path2, num_epochs=3):
    # Merge the two parquet files
    df1 = pd.read_parquet(data_path1)
    df2 = pd.read_parquet(data_path2)
    merged_df = pd.concat([df1, df2], ignore_index=True)

    # Create a new AIIAConfig instance
    config = AIIAConfig(
        model_name="AIIA-512x",
        hidden_size=512,
        num_hidden_layers=12,
        kernel_size=5,
        learning_rate=5e-5
    )

    # Initialize the base model
    model = AIIABase(config)

    # Create dataset loader with merged data
    train_dataset = AIIADataLoader(
        merged_df,
        batch_size=32,
        val_split=0.2,
        seed=42,
        column="image_bytes",
        label_column=None
    )

    # Create separate dataloaders for training and validation sets
    train_dataloader = DataLoader(
        train_dataset.train_dataset,
        batch_size=train_dataset.batch_size,
        shuffle=True,
        num_workers=4
    )

    val_dataloader = DataLoader(
        train_dataset.val_ataset,
        batch_size=train_dataset.batch_size,
        shuffle=False,
        num_workers=4
    )

    # Initialize loss functions and optimizer
    criterion_denoise = nn.MSELoss()
    criterion_rotate = nn.CrossEntropyLoss()

    optimizer = torch.optim.AdamW(model.parameters(), lr=config.learning_rate)

    device = "cuda" if torch.cuda.is_available() else "cpu"
    model.to(device)

    best_val_loss = float('inf')

    for epoch in range(num_epochs):
        print(f"\nEpoch {epoch+1}/{num_epochs}")
        print("-" * 20)
        
        # Training phase
        model.train()
        total_train_loss = 0.0

        for batch in train_dataloader:
            images, targets, tasks = zip(*batch)

            if device == "cuda":
                images = [img.cuda() for img in images]
                targets = [t.cuda() for t in targets]

            optimizer.zero_grad()

            # Process each sample individually since tasks can vary
            outputs = []
            total_loss = 0.0
            for i, (image, target, task) in enumerate(zip(images, targets, tasks)):
                output = model(image.unsqueeze(0))
                
                if task == 'denoise':
                    loss = criterion_denoise(output.squeeze(), target)
                elif task == 'rotate':
                    loss = criterion_rotate(output.view(-1, len(set(outputs))), target)
                
                total_loss += loss
                outputs.append(output)

            avg_loss = total_loss / len(images)
            avg_loss.backward()
            optimizer.step()

            total_train_loss += avg_loss.item()
            # Separate losses for reporting (you'd need to track this based on tasks)

        avg_total_train_loss = total_train_loss / len(train_dataloader)
        print(f"Training Loss: {avg_total_train_loss:.4f}")

        # Validation phase
        model.eval()
        with torch.no_grad():
            val_losses = []
            for batch in val_dataloader:
                images, targets, tasks = zip(*batch)

                if device == "cuda":
                    images = [img.cuda() for img in images]
                    targets = [t.cuda() for t in targets]

                outputs = []
                total_loss = 0.0
                for i, (image, target, task) in enumerate(zip(images, targets, tasks)):
                    output = model(image.unsqueeze(0))
                    
                    if task == 'denoise':
                        loss = criterion_denoise(output.squeeze(), target)
                    elif task == 'rotate':
                        loss = criterion_rotate(output.view(-1, len(set(outputs))), target)
                    
                    total_loss += loss
                    outputs.append(output)

                avg_val_loss = total_loss / len(images)
                val_losses.append(avg_val_loss.item())

            avg_val_loss = sum(val_losses) / len(val_dataloader)
            print(f"Validation Loss: {avg_val_loss:.4f}")

        # Save the best model
        if avg_val_loss < best_val_loss:
            best_val_loss = avg_val_loss
            model.save("BASEv0.1")
            print("Best model saved!")

if __name__ == "__main__":
    data_path1 = "/root/training_data/vision-dataset/images_checkpoint.parquet"
    data_path2 = "/root/training_data/vision-dataset/vec_images_dataset.parquet"
    pretrain_model(data_path1, data_path2, num_epochs=8)